/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/callback/OperationCallbackManager.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc. 
 * All rights reserved.
 */
package com.netapp.common.callback;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ConfigBundle;

/**
 * The OperationCallbackManager allows a client application to register a callback object to invoke
 * for a given operation (operationId or clientRequestId). The manager will retreive data for all
 * operations on a given server at once. The results of each call are then passed back to the callback
 * object so that the client application can access the changes.
 * <br />
 * Invoking the callback method will occur in a different thread. The results are queued up
 * on that thread in order to be processed.
 * 
 * @author joshb
 *
 */
public class OperationCallbackManager {
    private static final NALogger LOGGER = NALogger.getLogger(OperationCallbackManager.class);

    private static final Map<ServerEndPoint, CallbackPollThread> SERVER_CALLBACK_THREADS = new ConcurrentHashMap<ServerEndPoint, CallbackPollThread>();

    /*
     * By using a class that we will instantiate a new thread against, it is easier to build tests that do not require
     * remote access to a server
     */
    private Class<? extends CallbackPollThread> pollThreadClass;
    private ConfigBundle configuration;
    private boolean autoRelease;

    public OperationCallbackManager(ConfigBundle config) {
        setPollThreadClass(CallbackPollThread.class);
        setConfiguration(config);
        setAutoRelease(false);
    }
    
    public boolean register(String id, OperationCallback callback, ServerEndPoint endpoint) {
        boolean registered = true;
        
        CallbackPollThread pollThread = SERVER_CALLBACK_THREADS.get(endpoint);
        
        if (null == pollThread || !pollThread.isAlive()) {
            try {
                pollThread = getPollThreadClass().newInstance();
                pollThread.setEndpoint(endpoint);
                pollThread.setConfiguration(getConfiguration());
                pollThread.setAutoRelease(isAutoRelease());
                pollThread.start();
                SERVER_CALLBACK_THREADS.put(endpoint, pollThread);
            }
            catch (InstantiationException e) {
                LOGGER.error(MsgKey.CALLBACK_MANAGER_POLL_THREAD_CREATE_FAILURE, e, getPollThreadClass().getName());
                registered = false;
            }
            catch (IllegalAccessException e) {
                LOGGER.error(MsgKey.CALLBACK_MANAGER_POLL_THREAD_CREATE_FAILURE, e, getPollThreadClass().getName());
                registered = false;
            }
        }

        if (registered) {
            pollThread.addCallbackId(id, callback);
            LOGGER.debug(MsgKey.CALLBACK_MANAGER_REGISTERING_OPERATION, id, endpoint.getHostname());
        }
        
        return registered;
    }

    public void deregister(String id) {
        /*
         * Brute force attempt to just remove the id on every thread that is available
         */
        for (CallbackPollThread pollThread : SERVER_CALLBACK_THREADS.values()) {
            pollThread.removeCallbackEntry(id);
        }
    }

    /**
     * Remove all registered entries and stop any polling events that are ongoing
     */
    public void clear() {
        // stop polling threads
        for (CallbackPollThread thread : SERVER_CALLBACK_THREADS.values()) {
            thread.setStopped(true);
        }

        // Attempt to wait for each thread to finish
        for (CallbackPollThread thread : SERVER_CALLBACK_THREADS.values()) {
            try {
                thread.join();
            }
            catch (InterruptedException e) {
            }

            LOGGER.debug(MsgKey.CALLBACK_MANAGER_CLEARED_SERVER, thread.getCallbackIds().size(), thread.getEndpoint()
                .getHostname());
        }
        
        SERVER_CALLBACK_THREADS.clear();
    }

    protected Class<? extends CallbackPollThread> getPollThreadClass() {
        return pollThreadClass;
    }

    public void setPollThreadClass(Class<? extends CallbackPollThread> pollThreadClass) {
        this.pollThreadClass = pollThreadClass;
    }

    protected ConfigBundle getConfiguration() {
        return configuration;
    }

    protected void setConfiguration(ConfigBundle configuration) {
        this.configuration = configuration;
    }

    protected boolean isAutoRelease() {
        return autoRelease;
    }

    public void setAutoRelease(boolean autoRelease) {
        this.autoRelease = autoRelease;
    }
    
}
