/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/credentials/client/CredentialClient.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc. 
 * All rights reserved.
 */
package com.netapp.common.credentials.client;

import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.List;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;

import com.netapp.common.MsgKey;
import com.netapp.common.credentials.Credential;
import com.netapp.common.credentials.CredentialKey;
import com.netapp.common.credentials.service.CredentialService;
import com.netapp.common.credentials.service.CredentialServiceConstants;
import com.netapp.common.encryption.Encryption;
import com.netapp.common.encryption.EncryptionException;
import com.netapp.common.flow.api.OperationResult;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.StringUtils;

/**
 * Provides the plumbing for any external application to make the credential 
 * service SOAP calls.
 * 
 * The caller is responsible for ensuring that the context and credentialService have
 * been set prior to calling the init method. If either of these are null when init is
 * called, a RuntimeException will be thrown.
 * 
 * @author joshb
 *
 */
public class CredentialClient implements CredentialService {
    private static final NALogger LOGGER = NALogger.getLogger(CredentialClient.class);

    protected ConfigBundle context;
    protected PublicKey serverPublicKey;
    protected PrivateKey clientPrivateKey;
    protected CredentialService credentialService;
    protected String serverAlgorithm;
    protected String clientAlgorithm;

    public CredentialClient() {
    }

    /** 
     * Init method. To be called after all configuration is complete, generally
     * by Spring or another framework.
     *
     */
    public void init() {
        if (null != getContext()) {
            String publicKeyPath = CredentialClientConstants.SERVER_PUBLIC_KEY_PATH.lookup(getContext());
            String privateKeyPath = CredentialClientConstants.CLIENT_PRIVATE_KEY_PATH.lookup(getContext());
            String serverEncryptionAlgorithm = CredentialServiceConstants.SERVER_ENCRYPTION_ALGORITHM
                .lookup(getContext());
            String clientEncryptionAlgorithm = CredentialServiceConstants.CLIENT_ENCRYPTION_ALGORITHM
                .lookup(getContext());

            String errorMessage = "";
            String lineSeparator = System.getProperty("line.separator");

            if (StringUtils.isEmpty(publicKeyPath)) {
                errorMessage += LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_PROPERTY_IS_EMPTY,
                    CredentialClientConstants.SERVER_PUBLIC_KEY_PATH.getProperty())
                    + lineSeparator;
            }
            if (StringUtils.isEmpty(privateKeyPath)) {
                errorMessage += LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_PROPERTY_IS_EMPTY,
                    CredentialClientConstants.CLIENT_PRIVATE_KEY_PATH.getProperty())
                    + lineSeparator;
            }
            if (StringUtils.isEmpty(serverEncryptionAlgorithm)) {
                errorMessage += LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_PROPERTY_IS_EMPTY,
                    CredentialServiceConstants.SERVER_ENCRYPTION_ALGORITHM.getProperty())
                    + lineSeparator;
            }
            if (StringUtils.isEmpty(clientEncryptionAlgorithm)) {
                errorMessage += LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_PROPERTY_IS_EMPTY,
                    CredentialServiceConstants.CLIENT_ENCRYPTION_ALGORITHM.getProperty())
                    + lineSeparator;
            }
            
            if (!StringUtils.isEmpty(errorMessage)) {
                throw new RuntimeException(errorMessage);
            }

            try {
                setServerAlgorithm(serverEncryptionAlgorithm);
                setServerPublicKey(Encryption.getPublicKeyFromFile(publicKeyPath));
            }
            catch (EncryptionException e) {
                throw new RuntimeException(LOGGER.error(
                    MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_LOADING_PUBLIC_KEY, e, publicKeyPath,
                    getServerAlgorithm()));
            }

            try {
                setClientAlgorithm(clientEncryptionAlgorithm);
                setClientPrivateKey(Encryption.getPrivateKeyFromFile(privateKeyPath, getClientAlgorithm()));
            }
            catch (EncryptionException e) {
                throw new RuntimeException(LOGGER.error(
                    MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_LOADING_PRIVATE_KEY, e, privateKeyPath,
                    getClientAlgorithm()));
            }
        }
        else {
            throw new RuntimeException(LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_CONTEXT_NOT_SET));
        }

        if (null == getCredentialService()) {
            throw new RuntimeException(LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_CREDENTIAL_SERVICE_NOT_SET));
        }
    }

    /*
     * The following methods will make the SOAP call to a remote server.
     */

    /**
     * Encrypts the password using the configured public key before making the SOAP call.
     *
     * @throws RuntimeException
     */
    public OperationResult addCredential(Credential credential, String type) {
        try {
            byte[] encryptedPassword = Encryption.encrypt(credential.getPassword().getBytes(), getServerPublicKey(),
                getServerAlgorithm());
            String hexEncoded = new String(Hex.encodeHex(encryptedPassword));

            Credential secureCredential = new Credential(credential.getKey(), hexEncoded);
            return getCredentialService().addCredential(secureCredential, type);
        }
        catch (EncryptionException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Retrieves the credential from the remote server and decrypts the password using
     * the configured public key
     * 
     * @throws RuntimeException
     */
    public Credential getCredential(CredentialKey key, String type) {
        try {
            Credential secureCredential = getCredentialService().getCredential(key, type);

            if (null == secureCredential) {
                return null;
            }

            byte[] hexString = Hex.decodeHex(secureCredential.getPassword().toCharArray());

            byte[] decryptedPassword = Encryption.decrypt(hexString, getClientPrivateKey(), getClientAlgorithm());

            return new Credential(secureCredential.getKey(), new String(decryptedPassword));
        }
        catch (EncryptionException e) {
            throw new RuntimeException(e);
        }
        catch (DecoderException e) {
            throw new RuntimeException(e);
        }
    }

    public OperationResult removeCredential(CredentialKey key, String type) {
        return getCredentialService().removeCredential(key, type);
    }
    
    /**
     * Currently, listing credentials does not pass the password. If this changes,
     * then the passwords will need to be decrypted exactly the same as getCredential does.
     */
    public List<Credential> listCredentials(String type) {
        return getCredentialService().listCredentials(type);
    }

    public String interfaceVersion() {
        return getCredentialService().interfaceVersion();
    }

    public String releaseVersion() {
        return getCredentialService().releaseVersion();
    }

    /*
     * Bean properties for member variables. These are protected unless required to be set by external callers.
     */

    protected ConfigBundle getContext() {
        return context;
    }

    public void setContext(ConfigBundle context) {
        this.context = context;
    }

    protected PublicKey getServerPublicKey() {
        return serverPublicKey;
    }

    protected void setServerPublicKey(PublicKey serverPublicKey) {
        this.serverPublicKey = serverPublicKey;
    }

    protected CredentialService getCredentialService() {
        return credentialService;
    }

    public void setCredentialService(CredentialService credentialService) {
        this.credentialService = credentialService;
    }

    protected String getServerAlgorithm() {
        return serverAlgorithm;
    }

    protected void setServerAlgorithm(String serverAlgorithm) {
        this.serverAlgorithm = serverAlgorithm;
    }

    protected PrivateKey getClientPrivateKey() {
        return clientPrivateKey;
    }

    protected void setClientPrivateKey(PrivateKey clientPrivateKey) {
        this.clientPrivateKey = clientPrivateKey;
    }

    protected String getClientAlgorithm() {
        return clientAlgorithm;
    }

    protected void setClientAlgorithm(String clientAlgorithm) {
        this.clientAlgorithm = clientAlgorithm;
    }

}
