/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/EngineConfiguration.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import static com.netapp.common.flow.FlowConstants.DIAGNOSTIC_OPERATION_NAME;
import static com.netapp.common.flow.FlowConstants.OPERATION_NAME_REGEX;
import static com.netapp.common.flow.FlowConstants.WORKFLOW_CATALOG_PATH;
import static com.netapp.common.flow.FlowConstants.WORKFLOW_SCHEMA_URI;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.xml.CatalogEntityResolver;
import com.netapp.common.xml.DefaultErrorHandler;

/**
 * Container for the configuration data for the workflow engine.
 * Exposed so we can configure via Spring.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class EngineConfiguration {
    private static final NALogger logger = NALogger.getLogger(EngineConfiguration.class);
    
    private Map<String, OperationTemplate> operations = new HashMap<String, OperationTemplate>();
    private DiagnosticTemplate diagnosticTemplate;
    
    private DocumentBuilderFactory docFactory;
    private CatalogEntityResolver entityResolver;
    private TaskInstanceFactory taskInstanceFactory;
    
    /**
     * Construct a new empty EngineConfiguration
     *
     */
    public EngineConfiguration() {
        // TODO: Fix this before checkin.. We need to resolve these another way
        this.entityResolver = new CatalogEntityResolver();
        this.entityResolver.addCatalogs(WORKFLOW_CATALOG_PATH);
        
        this.docFactory = DocumentBuilderFactory.newInstance();
        this.docFactory.setNamespaceAware(true);
        this.docFactory.setXIncludeAware(true);
        this.docFactory.setAttribute("http://java.sun.com/xml/jaxp/properties/schemaLanguage", XMLConstants.W3C_XML_SCHEMA_NS_URI);
        this.docFactory.setAttribute("http://java.sun.com/xml/jaxp/properties/schemaSource", WORKFLOW_SCHEMA_URI);
        this.docFactory.setValidating(true);
        this.docFactory.setIgnoringComments(true);
        // This doesn't actually work
        // http://forum.java.sun.com/thread.jsp?forum=34&thread=430252
        // http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4867706
        // this.docFactory.setIgnoringElementContentWhitespace(true);
        
        this.taskInstanceFactory = new DefaultTaskInstanceFactory();
    }
    
    public void addOperation(String name, String resource) {
        InputStream is = Thread.currentThread().getContextClassLoader().getResourceAsStream(resource);
        if (is == null) {
            throw new FlowException(logger.error(MsgKey.FLOW_CONFIGURATION_FILE_NOT_FOUND, resource));
        }
        addOperation(name, is);
    }
    
    public void addOperation(String name, InputStream stream) {
        if (this.operations.containsKey(name)) {
            throw new FlowException(logger.error(MsgKey.FLOW_CONFIGURATION_OPERATION_ALREADY_DEFINED, name));
        }

        logger.internal("Defining operation " + name);
        Document doc = null;
        try {
            DocumentBuilder builder = getDocumentBuilder();
            DefaultErrorHandler eh = new DefaultErrorHandler();
            builder.setErrorHandler(eh);
            doc = builder.parse(stream);
            if (eh.hasErrors()) {
                for (SAXParseException e : eh.getErrors()) {
                    logger.error(MsgKey.FLOW_CONFIGURATION_PARSE_EXCEPTION, e.getLineNumber(), e.getMessage());
                }
                throw new FlowException(logger.error(MsgKey.FLOW_CONFIGURATION_ERROR));
            }
            Node root = doc.getDocumentElement();

            if (DIAGNOSTIC_OPERATION_NAME.equals(name)) {
                this.diagnosticTemplate = new DiagnosticTemplate(root, this.taskInstanceFactory);
            }
            else if (OPERATION_NAME_REGEX.matcher(name).matches()) {
                OperationTemplate operation = new OperationTemplate(name, root, this.taskInstanceFactory);
                this.operations.put(name, operation);
            }
            else {
                throw new FlowException(logger.error(MsgKey.FLOW_INVALID_OPERATION_NAME_DEFINED, name));
            }
        }
        catch (ParserConfigurationException e) {
            throw new RuntimeException(e);
        }
        catch (SAXException e) {
            throw  new RuntimeException(e);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
    
    /**
     * Get a list of configured operations this engine has
     * @return Configured operations
     */
    List<OperationInformation> getRegisteredOperations() {
        List<OperationInformation> infos = new ArrayList<OperationInformation>();
        for (OperationTemplate operation : operations.values()) {
            infos.add(operation.getInformation());
        }
        return infos;
    }
    
    Collection<OperationTemplate> getOperationTemplates() {
        List<OperationTemplate> list = new ArrayList<OperationTemplate>();
        list.addAll(this.operations.values());
        if (this.diagnosticTemplate != null) {
            list.add(this.diagnosticTemplate);
        }
        return list;
    }
    
    /**
     * 
     */
    OperationTemplate getOperationTemplate(String operationName) {
        if (operationName.equals(DIAGNOSTIC_OPERATION_NAME)) {
            return this.diagnosticTemplate;
        }
        
        return operations.get(operationName);
    }
    
    DiagnosticTemplate getDiagnosticTemplate() {
        return diagnosticTemplate;
    }
    
    private synchronized DocumentBuilder getDocumentBuilder() throws ParserConfigurationException {
        DocumentBuilder builder = this.docFactory.newDocumentBuilder();
        if (this.entityResolver != null) {
            builder.setEntityResolver(entityResolver);
        }
        return builder;
    }
    
    public void setOperations(Set<OperationResource> operationResources) {
        for (OperationResource resource : operationResources) {
            addOperation(resource.getOperationName(), resource.getWorkflowResource());
        }
    }
    
    public void setTaskInstanceFactory(TaskInstanceFactory taskInstanceFactory) {
        this.taskInstanceFactory = taskInstanceFactory;
    }
}
