/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/PersistenceManager.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import java.util.List;
import java.util.Map;
import java.util.Set;

import com.netapp.common.logging.LogMessage;

/**
 * Interface for classes responsible for creating and persisting 
 * {@link com.netapp.common.flow.OperationContext} objects.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public interface PersistenceManager {
    
    /**
     * Create a new OperationContext managed by this manager.
     * 
     * @param template Operation Template for new context
     * @param params Initial register values for the new context
     * @return New operation context
     */
    OperationContext createContext(OperationTemplate template, Map<String, ?> params);

    /**
     * Get the context by the supplied id.
     * If id is a 32 bit guid, the operation will be looked up by the primary id, else
     * the id will be assumed to be a client request id.
     * @param id
     * @return Persisted context with the supplied id, or null if the supplied id does not exist.
     */
    OperationContext getContext(String id);
    
    /**
     * Get the next context available to run from the given list
     * of operation types.
     * 
     * @param operationTypes
     * @return Persisted context to begin executing. The context will have already been
     * transistioned to the RUNNING state. If no contexts are available to run immediately,
     * null is returned.
     */
    OperationContext getRunnableContext(Set<String> operationTypes);
    
    /**
     * Save (INSERT or UPDATE) the supplied context. Additionally, if the
     * context is in a terminal state, any locks held by the context should be released.
     * @param context
     */
    void saveContext(OperationContext context);
    
    /**
     * Attempt to cancel the requested operation.
     * @param id
     */
    void cancelContext(String id);
    
    /**
     * Inform the PersistenceManager the supplied ID is no longer required.
     * If the operation is in a terminal state, the manager may stop holding
     * information for the supplied context.
     * 
     * This is an optional method and is not required to be implemented by a
     * manager, in which case it should just be a no-op.
     * @param id
     */
    void releaseContext(String id);
    
    /**
     * Log a message against the specified context.
     * @param id Id of the operation message was logged under
     * @param message message details
     */
    void addMessage(OperationId id, LogMessage message);

    /**
     * Get the list of messages logged against a specific context.
     * @param id
     * @return List of log messages
     */
    List<LogMessage> getMessages(OperationId id);

    /**
     * Acquire a list of locks on the supplied string. Only one operation
     * may hold a lock at any given time, once acquired, locks are
     * freed automatically at the end of an operation.
     * The same context may acquire a lock multiple times without
     * penalty. Attempting to acquire a lock held by a different transaction
     * will throw a UnableToAcquireLockException.
     * 
     * @param id
     * @param keys
     */
    void acquireLocks(OperationId id, Set<String> keys) throws UnableToAcquireLockException;
    
    /**
     * Start any final resources required for this manager.
     */
    void start(Engine engine);
    
    /**
     * Stop any resources used by this manager.
     */
    void stop();
    
    boolean isValid();
}
