/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/logging/LogMessage.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.logging;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.regex.Pattern;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlType;

import com.netapp.common.util.StringUtils;

/**
 * A localized log message suitable for serializing to a remote client. Also includes routines
 * for relocalizing on a client.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
@XmlType(name = "logMessage", namespace = "http://www.netapp.com/smai/Operation")
@XmlAccessorType(XmlAccessType.FIELD)
public class LogMessage implements Serializable {
    private static final String BOOLEAN_REGEX = "(true|false)";
    private static final String INT_REGEX = "[\\-\\+]?\\d+";
    private static final String FLOAT_REGEX = "[\\-\\+]?\\d?\\.\\d+";
    private static final long serialVersionUID = 0L;

    @XmlElement
    protected Date timestamp;
    @XmlElement
    protected LogLevel level;
    @XmlElement
    protected String thread;
    @XmlElement
    protected String location;
    
    @XmlElement
    protected String msgKeyClass;
    @XmlElement
    protected String msgKeyValue;
    protected transient MessageKey msgKey;
    
    @XmlElementWrapper(name = "parameters", nillable = true)
    @XmlElement(name = "parameter")
    List<String> parameters;
    
    protected transient Object[] castParameters;
    
    @XmlElement
    protected String message;
    protected transient String localizedMessage;

    /**
     * Construct an empty log message for deserialization
     *
     */
    public LogMessage() {
        this.parameters = new ArrayList<String>();
    }

    public LogMessage(LogLevel level, String location, MessageKey msgKey, Object... parameters) {
        this.timestamp = new Date();
        this.level = level;
        this.thread = Thread.currentThread().getName();
        this.location = location;

        this.msgKey = msgKey;
        this.msgKeyClass = msgKey.getClass().getName();
        this.msgKeyValue = msgKey.name();

        if (parameters != null && parameters.length > 0) {
            this.message = MessageBundle.localizeMessage(msgKey, parameters);
            this.parameters = new ArrayList<String>();
            for (int i = 0; i < parameters.length; i++) {
                String parameter = (parameters[i] == null) ? "null" : parameters[i].toString();
                this.parameters.add(parameter);
            }
        }
        else {
            this.message = MessageBundle.localizeMessage(msgKey);
        }
        this.localizedMessage = this.message;
    }

    /**
     * Set the timestamp for this log message
     * @param timestamp Time this message was logged, in milliseconds since epoch
     */
    public void setTimestamp(Date timestamp) {
        assert this.timestamp == null;
        this.timestamp = timestamp;
    }

    /**
     * Retrieve the timestamp for this log message
     * @return Time this message was logged, in milliseconds since epoch
     */
    public Date getTimestamp() {
        return this.timestamp;
    }

    public void setLevel(LogLevel level) {
        assert this.level == null;
        this.level = level;
    }

    /**
     * Retrieve the level this message was logged at.
     * @return log level of this message
     */
    public LogLevel getLevel() {
        return this.level;
    }

    /**
     * Set the location (logger) this message was logged at. Generally this is
     * the class where NALogger.log() was called.
     * @param location
     */
    public void setLocation(String location) {
        this.location = location;
    }

    /**
     * Retrieve the location (logger) this message was logged at. Generally this
     * is the class where NALogger.log() was called.
     * @return logger that logged this message
     */
    public String getLocation() {
        return this.location;
    }

    public void setThread(String thread) {
        this.thread = thread;
    }

    /**
     * The name of the thread where this message was logged.
     * @return thread that logged this message
     */
    public String getThread() {
        return this.thread;
    }

    public void setMessage(String message) {
        assert this.message == null;
        this.message = message;
    }
    /**
     * The default message text for this message. Generally this is the localized version
     * from the system that first emitted the message.
     * @return message text
     */
    public String getMessage() {
        return this.message;
    }

    /**
     * The locally (client) localized version of the message, or the same as {@link #getMessage()} if
     * the system is unable to re-localize this message.
     * @return message text
     */
    public String localizedMessage() {
        if (this.localizedMessage == null) {
            if (this.msgKey() != null) {
                if (this.castParameters() != null) {
                    this.localizedMessage = MessageBundle.localizeMessage(this.msgKey(), this.castParameters());
                }
                else {
                    this.localizedMessage = MessageBundle.localizeMessage(this.msgKey());
                }
            }
            // If we couldn't regenerate the MsgKey, use the unlocalized version
            else {
                this.localizedMessage = this.message;
            }
        }
        return this.localizedMessage;
    }

    public void setMsgKeyClass(String messageKeyClass) {
        assert this.msgKeyClass == null;
        this.msgKeyClass = messageKeyClass;
    }

    /**
     * The Java class of the localized message key for this message
     * @return localized message key class
     */
    public String getMsgKeyClass() {
        return this.msgKeyClass;
    }

    public void setMsgKeyValue(String msgKeyValue) {
        assert this.msgKeyValue == null;
        this.msgKeyValue = msgKeyValue;
    }

    /**
     * The enum value of the localized message key for this message.
     * @return localized message key enum
     */
    public String getMsgKeyValue() {
        return this.msgKeyValue;
    }

    public void setParameters(List<String> parameters) {
        this.parameters = parameters;
    }

    /**
     * Substituted parameters for this message
     * @return message parameters
     */
    public List<String> getParameters() {
        return this.parameters;
    }

    /**
     * Relog this message through the system logger
     * and any associated listeners. The log will be relogged
     * at the original level and category.
     *
     */
    public void relog() {
        NALogger.getLogger(this.location).log(this);
    }

    /**
     * Retrieve the original MsgKey for this message, or null if the
     * key can not be recreated.
     * @return key of the original message
     */
    @SuppressWarnings("unchecked")
    public MessageKey msgKey() {
        if (this.msgKey == null 
                && !StringUtils.isEmpty(this.msgKeyClass) 
                && !StringUtils.isEmpty(this.msgKeyValue)) {
            try {
                Class<?> enumClass = Class.forName(this.msgKeyClass);
                if (enumClass.isEnum() && MessageKey.class.isAssignableFrom(enumClass)) {
                    this.msgKey = (MessageKey) Enum.valueOf(enumClass.asSubclass(Enum.class), this.msgKeyValue);
                }
            }
            catch (ClassNotFoundException e) {
                // We couldn't find the message key
            }
        }
        return this.msgKey;
    }

    /**
     * Convert the supplied string parameters into primitives suitable for use with
     * sprintf style format specifiers.
     * @return The parameters for this message key. Numbers are cast into a suitable type, other values are returned as strings.
     */
    public Object[] castParameters() {
        if (this.castParameters == null && this.parameters != null && this.parameters.size() > 0) {
            this.castParameters = new Object[this.parameters.size()];
            for (int i = 0; i < this.parameters.size(); i++) {
                String parameter = this.parameters.get(i);
                if (Pattern.matches(BOOLEAN_REGEX, parameter)) {
                    this.castParameters[i] = Boolean.parseBoolean(parameter);
                }
                else if (Pattern.matches(INT_REGEX, parameter)) {
                    this.castParameters[i] = Long.parseLong(parameter);
                }
                else if (Pattern.matches(FLOAT_REGEX, parameter)) {
                    this.castParameters[i] = Double.parseDouble(parameter);
                }
                else {
                    this.castParameters[i] = parameter;
                }
            }
        }
        return this.castParameters;
    }

    @Override
    public String toString() {
        return this.message;
    }
}

