/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/server/ServerDescriptor.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.server;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;

/**
 * Wraps the description of a running process. A process runs in a PID on a PORT. A descriptor
 * can be created to contain this information. This is useful for a) ensuring only one process
 * is running on the server, b) finding out the PID in order to issue a kill command to stop the process.
 *
 */
public class ServerDescriptor {

    private static final NALogger LOGGER = NALogger.getLogger(ServerDescriptor.class);
    public static final File DEFAULT_PID_DIR = new File("/var/run");
    
    private static final String PID_FILE_REGEX = "[0-9]+\\.pid";
    private static final String PORT_PROPERTY = "port";
    private static final String PID_PROPERTY = "pid";
    private static final String INSTALL_DIR = "install.dir";
    
    private int pid;
    private int port;
    private String dir;
    private String prefix;
    private String installDirectory;
    private String pidDirectory;
    
    public ServerDescriptor(String prefix, int pid, int port) {
        // called by startServer to create a file
        this.pid = pid;
        this.port = port;
        setPrefix(prefix);
        setPidDirectory(DEFAULT_PID_DIR.getPath());
    }
        
    protected ServerDescriptor(File descriptorFile) throws FileNotFoundException {
        Properties properties = new Properties();
        try {
            properties.load(new FileInputStream(descriptorFile));
            this.pid = Integer.parseInt(properties.getProperty(PID_PROPERTY));
            this.port = Integer.parseInt(properties.getProperty(PORT_PROPERTY));
            this.dir = properties.getProperty(INSTALL_DIR);
        }
        catch (FileNotFoundException e) {
            throw e;
        }
        catch (IOException e) {
            throw new ServerException(MessageBundle.localizeMessage(
                MsgKey.SERVER_BASE_CANNOT_READ_DESCRIPTOR_FILE, descriptorFile, e.getLocalizedMessage()), e);            
        }
    }
    
    public void toFile() {
        File descriptorFile = getFile(this.pid);

        try {
            Properties properties = new Properties();
            properties.setProperty(PID_PROPERTY, Integer.toString(this.pid));
            properties.setProperty(PORT_PROPERTY, Integer.toString(this.port));
            
            if (null != getInstallDirectory() && 0 < getInstallDirectory().length()) {
                File installDir = new File(getInstallDirectory());
                properties.setProperty(INSTALL_DIR, installDir.getCanonicalPath());
            }
            
            properties.setProperty("user.dir", System.getProperty("user.dir"));
            properties.setProperty("java.home", System.getProperty("java.home"));
            // properties.setProperty("version", SMOVersion.VERSION);
            properties.store(new FileOutputStream(descriptorFile), "Server Descriptor File");
        }
        catch (IOException e) {
            throw new ServerException(MessageBundle.localizeMessage(
                MsgKey.SERVER_BASE_CANNOT_WRITE_DESCRIPTOR_FILE, descriptorFile, e.getLocalizedMessage()), e);
        }
    }
    
    public boolean deleteFile() {
        File f = getFile(this.pid);
        return f.delete();
    }
    
    public int getPid() {
        return this.pid;
    }
    
    public int getPort() {
        return this.port;
    }
    
    public String getDir() {
        return this.dir;
    }
    
    public String getPrefix() {
        return prefix;
    }

    public void setPrefix(String prefix) {
        this.prefix = prefix;
    }
    
    public String getInstallDirectory() {
        return installDirectory;
    }

    public void setInstallDirectory(String installDirectory) {
        this.installDirectory = installDirectory;
    }

    public String getPidDirectory() {
        return pidDirectory;
    }

    public void setPidDirectory(String pidDirectory) {
        this.pidDirectory = pidDirectory;
        setupDirectoryForPidFile(new File(this.pidDirectory));
    }

    public String getFileName() {
        return getFileName(this.pid);
    }
        
    public String getFileName(int pid) {
        return getPrefix() + pid + ".pid";
    }
    
    public File getFile(int pid) {
        String descriptorFileName = getFileName(pid);
        return new File(getPidDirectory(), descriptorFileName);
    }
    
    private File[] getAllFiles() {
        File varDir = new File(getPidDirectory());
        File[] files = varDir.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.matches(getRegex());
            }
        });
        return files;
    }
    
    public List<ServerDescriptor> getAllDescriptors() {
        File[] allFiles = getAllFiles();
        List<ServerDescriptor> result = new ArrayList<ServerDescriptor>();
        for (File file : allFiles) {
            try {
                ServerDescriptor descriptor = new ServerDescriptor(file);
                
                descriptor.setInstallDirectory(getInstallDirectory());
                descriptor.setPidDirectory(getPidDirectory());
                descriptor.setPrefix(getPrefix());
                
                result.add(descriptor);
            }
            catch (FileNotFoundException fnfe) {
                // skip it
            }
        }
        return result;
    }

    protected String getRegex() {
        return getPrefix() + PID_FILE_REGEX;
    }

    private File setupDirectoryForPidFile(File pidDirectory) {
        if (!pidDirectory.exists()) {
            pidDirectory.mkdirs();
        }
        
        return pidDirectory;
    }
    
    @Override
    public String toString() {
        return " pid " + this.pid + " port " + this.port + " dir " + this.dir;
    }
}
