/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/servlet/wsdldoc/WsdlDocumentationServlet.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.servlet.wsdldoc;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
/**
 * This servlet will transform the WSDL for the requested service using 
 * an XSL file specified by the init-param wsdlDocXSL.  The XSL file is 
 * loaded from the servlet's classpath.  
 * 
 * The WSDL is retrieved from the URL specified by the init-param wsdlURL
 * with the service name and ?WSDL appended to the URL.
 * 
 * The service name is determined from the incoming request's
 * servlet path info.
 * 
 * A sample web.xml configuration looks like:
 *   &lt;servlet&gt;
 *   &lt;servlet-name&gt;wsdlDoc&lt;/servlet-name&gt;
 *   &lt;servlet-class&gt;com.netapp.common.servlet.wsdldoc.WsdlDocumentationServlet&lt;/servlet-class&gt;
 *   &lt;init-param&gt;
 *       &lt;param-name&gt;wsdlDocXSL&lt;/param-name&gt;
 *       &lt;param-value&gt;wsdl-viewer.xsl&lt;/param-value&gt;
 *   &lt;/init-param&gt;
 *   &lt;init-param&gt;
 *       &lt;param-name&gt;wsdlURL&lt;/param-name&gt;
 *       &lt;param-value&gt;../services&lt;/param-value&gt;
 *   &lt;/init-param&gt;
 *   &lt;init-param&gt;
 *       &lt;param-name&gt;contextConfigLocation&lt;/param-name&gt;
 *       &lt;param-value&gt;
 *         classpath:configuration/spring/workflow.xml 
 *         classpath:configuration/spring/csm-spring.xml 
 *         classpath:org/codehaus/xfire/spring/xfire.xml
 *       &lt;/param-value&gt;
 *   &lt;/init-param&gt;
 * &lt;/servlet&gt;
 *
 * @author odom
 *
 */
public class WsdlDocumentationServlet extends HttpServlet {
    private static final String ALGORITHM_SSL = "ssl";
    private static final String WSDL_QUERY_PARAMETER = "?WSDL";
    private static final String PROTOCOL_HTTPS = "https";
    
    private static final long serialVersionUID = 1L;

    /**
     * A dummy X509TrustManager that will blindly accept any certificate.
     * @author odom
     *
     */
    private class SilentX509TrustManager implements X509TrustManager {
        public X509Certificate[] getAcceptedIssuers() {
            return null;
        }

        public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException {
        }

        public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException {
        }
    }    
    
    /**
     * A dummy HostnameVerifier that will blindly accept any host name.
     * @author odom
     *
     */
    private class SilentHostnameVerifier implements HostnameVerifier {
        public boolean verify(String s, SSLSession sslsession) {
            return true;
        }
        
    }
    
    private TransformerFactory transformerFactory = TransformerFactory.newInstance();
    private SSLSocketFactory veryTrustingSocketFactory;
    private String wsdlURL;
    private String wsdlDocXSL;
        
    @Override
    public void init() throws ServletException {
        initSSL();
        this.wsdlURL = getServletConfig().getInitParameter("wsdlURL");
        this.wsdlDocXSL = getServletConfig().getInitParameter("wsdlDocXSL");
    }
    
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException {
        try {            
            transformWSDL(fetchWSDL(new URL(request.getRequestURL().toString()),
                this.wsdlURL + request.getPathInfo() + WSDL_QUERY_PARAMETER), 
                response.getOutputStream());
        }
        catch (TransformerException e) {
            throw new ServletException(e.getMessage(), e);
        }
        catch (MalformedURLException e) {
            throw new ServletException(e.getMessage(), e);
        }
        catch (IOException e) {
            throw new ServletException(e.getMessage(), e);
        }        
    }
    
    private InputStream fetchWSDL(URL contextURL, String wsdlRequestURL) throws IOException {
        URL requestURL = new URL(contextURL, wsdlRequestURL);
        URLConnection connection = createURLConnection(requestURL);
        connection.connect();
        return connection.getInputStream();
    }
    
    private URLConnection createURLConnection(URL requestURL) throws IOException {
        URLConnection urlConnection = requestURL.openConnection();
        
        if (urlConnection instanceof HttpsURLConnection) {            
            ((HttpsURLConnection)urlConnection).setSSLSocketFactory(this.veryTrustingSocketFactory);
            ((HttpsURLConnection)urlConnection).setHostnameVerifier(new SilentHostnameVerifier());
        }
        
        return urlConnection;
    }
    
    private InputStream fetchXSL() {
        return this.getClass().getClassLoader().getResourceAsStream(this.wsdlDocXSL);
    }
    
    private void transformWSDL(InputStream wsdlStream, OutputStream outputStream) throws TransformerException {
        Transformer xslTransformer = this.transformerFactory.newTransformer(new StreamSource(fetchXSL()));
        xslTransformer.transform(new StreamSource(wsdlStream), new StreamResult(outputStream));
    }

    private void initSSL() throws ServletException {
        
        try {
            SSLContext sslContext = SSLContext.getInstance(ALGORITHM_SSL);
            sslContext.init(null, new TrustManager[]{new SilentX509TrustManager()}, new SecureRandom());
            veryTrustingSocketFactory = sslContext.getSocketFactory();
        }
        catch (NoSuchAlgorithmException e) {
            throw new ServletException(e.getMessage(), e);
        }
        catch (KeyManagementException e) {
            throw new ServletException(e.getMessage(), e);
        }
    }
    
    public String getWsdlURL() {
        return wsdlURL;
    }

    public void setWsdlURL(String wsdlURL) {
        this.wsdlURL = wsdlURL;
    }

    public String getWsdlDocXSL() {
        return wsdlDocXSL;
    }

    public void setWsdlDocXSL(String wsdlDocXSL) {
        this.wsdlDocXSL = wsdlDocXSL;
    }
}
