/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/spring/CachingComponentScanBeanDefinitionParser.java#1 $
 * 
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.spring;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Set;

import org.apache.commons.lang.time.StopWatch;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.BeanDefinitionHolder;
import org.springframework.beans.factory.xml.ParserContext;
import org.springframework.beans.factory.xml.XmlReaderContext;
import org.springframework.context.annotation.ComponentScanBeanDefinitionParser;
import org.springframework.core.io.FileSystemResource;
import org.w3c.dom.Element;

import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.FileSecurityDescriptor;
import com.netapp.common.platform.OperatingSystem;

public class CachingComponentScanBeanDefinitionParser extends ComponentScanBeanDefinitionParser {

    private static final NALogger logger = NALogger.getLogger();
    
    private static final String CACHE_KEY_ATTRIBUTE = "cache-key";
    
    private static final String ANNOTATION_CONFIG_ATTRIBUTE = "annotation-config";
    
    private static final String[] CANDIDATE_CACHE_PARENT_DIRS = new String[] { "tmp" }; 
    
    private static final File CACHE_DIR;

    private static final OperatingSystem OPERATING_SYSTEM = OperatingSystem.createInstance();
    
    static {
        File candidateCacheDir = null;
        
        for (int i = 0; i < CANDIDATE_CACHE_PARENT_DIRS.length; i++) {
            File candidateCacheParentDir = new File(CANDIDATE_CACHE_PARENT_DIRS[i]);
            
            if (candidateCacheParentDir.exists() 
                    && candidateCacheParentDir.isDirectory() 
                    && candidateCacheParentDir.canWrite()) {
            
                candidateCacheDir = new File(candidateCacheParentDir, "spring-cache");
                
                if (!candidateCacheDir.exists()) {
                    candidateCacheDir.mkdir();
                    makeReadWriteExecuteByAll(candidateCacheDir);
                    break;
                }
                else if (candidateCacheDir.canWrite()) {
                    break;    
                }
                else {
                    candidateCacheDir = null;
                }
            }
        }
        
        if (candidateCacheDir == null) {
            throw new IllegalStateException(
                String.format("Unable to create spring-cache directory in any of %s", Arrays.asList(CANDIDATE_CACHE_PARENT_DIRS)));
        }
        
        CACHE_DIR = candidateCacheDir;
        
        logger.internal("CachingComponentScanBeanDefinitionParser: Using directory %s", CACHE_DIR.getAbsolutePath());
    }
    
    /*
     * I really don't like having a dependency on the native libraries here, 
     * but in order for cache files created by this process to get cleaned
     * up during the build process, the files and directories created here need
     * to be r/w/x by all
     */
    
    private static void makeReadWriteExecuteByAll(File file) {
        // Skip this step if the cache is on a CIFS share and setFileSecurityDescriptor
        // doesn't work right
        if (System.getProperty("com.netapp.common.spring.cache.cifshack") != null) {
            return;
        }
        FileSecurityDescriptor fileSecurityDescriptor = OPERATING_SYSTEM.getFileSecurityDescriptor(file);
        fileSecurityDescriptor.addAccessControls(
            FileSecurityDescriptor.AccessControl.ALL_READ, 
            FileSecurityDescriptor.AccessControl.ALL_WRITE,
            FileSecurityDescriptor.AccessControl.ALL_EXECUTE
        );
        
        OPERATING_SYSTEM.changeFileSecurity(file, fileSecurityDescriptor);
    }
    
    public BeanDefinition parse(Element element, ParserContext parserContext) {
        
        if (CACHE_DIR == null) {
            // Caching is disabled
            return super.parse(element, parserContext);
        }
        
        String cacheKey = element.getAttribute(CACHE_KEY_ATTRIBUTE);
        
        File cacheFile = this.getCacheFile(cacheKey);
        
        StopWatch stopWatch = new StopWatch();
        stopWatch.start();
        
        boolean isCacheHit = cacheFile.exists();
        
        if (isCacheHit) {
            this.loadFromCache(cacheFile, parserContext);
        }
        else {
            super.parse(element, parserContext);
        }
        
        stopWatch.stop();
        logger.internal("CachingComponentScanBeanDefinitionParser: Cache[%s] %s (%s)%n", cacheKey, isCacheHit ? "HIT" : "MISS", stopWatch.toString());
        
        return null;
    }

    private File getCacheFile(String cacheKey) {
        return new File(CACHE_DIR, cacheKey + ".xml");
    }

    private void loadFromCache(File cacheFile, ParserContext parserContext) {
        parserContext.getReaderContext().getReader().loadBeanDefinitions(new FileSystemResource(cacheFile));
    }

    @Override
    protected void registerComponents(XmlReaderContext readerContext, Set<BeanDefinitionHolder> beanDefinitions, Element element) {
        super.registerComponents(readerContext, beanDefinitions, element);
        
        if (CACHE_DIR == null) {
            // Caching is disabled
            return;
        }
        
        String cacheKey = element.getAttribute(CACHE_KEY_ATTRIBUTE);
        
        File cacheFile = this.getCacheFile(cacheKey);
        
        PrintStream printStream = null;
        try {
            printStream = new PrintStream(new FileOutputStream(cacheFile));
            
            printStream.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            printStream.println();
            printStream.println("<beans xmlns=\"http://www.springframework.org/schema/beans\"");
            printStream.println("\txmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
            printStream.println("\txmlns:context=\"http://www.springframework.org/schema/context\"");
            printStream.println("\txsi:schemaLocation=\"http://www.springframework.org/schema/beans http://www.springframework.org/schema/beans/spring-beans-2.5.xsd");
            printStream.println("\thttp://www.springframework.org/schema/context http://www.springframework.org/schema/context/spring-context-2.5.xsd\"");
            printStream.println("\tdefault-lazy-init=\"true\">");
            
            printStream.println();
            
            // Register annotation config processors, if necessary.
            boolean annotationConfig = true;
            if (element.hasAttribute(ANNOTATION_CONFIG_ATTRIBUTE)) {
                annotationConfig = Boolean.valueOf(element.getAttribute(ANNOTATION_CONFIG_ATTRIBUTE));
            }
            
            if (annotationConfig) {
                printStream.printf("\t<context:annotation-config/>%n%n");
            }
            
            for (BeanDefinitionHolder beanDefinitionHolder : beanDefinitions) {
                String beanName = beanDefinitionHolder.getBeanName();
                
                /*
                 * The beanName must conform to the xsd:id datatype, which is
                 * defined in http://www.w3.org/XML/1998/namespace as:
                 * 
                 * The normalized value of the attribute is an NCName according to the 
                 * Namespaces in XML Recommendation which has the same version as the 
                 * document in which this attribute occurs (NCName for XML 1.0, or 
                 * NCName for XML 1.1).
                 * 
                 * The valid format of NCName is defined at http://www.w3.org/TR/xmlschema-2/#NCName
                 * NCName   ::=    (Letter | '_') (NCNameChar)*
                 * NCNameChar  ::=     Letter | Digit | '.' | '-' | '_' | CombiningChar | Extender
                 * 
                 * Basically, '$' is invalid, which can appear in beanNames generated from class names
                 * (for inner classes).  So replace it with '-'
                 */
                beanName = beanName.replace('$', '-');
                
                BeanDefinition beanDefinition = beanDefinitionHolder.getBeanDefinition();
                String beanClassName = beanDefinition.getBeanClassName();
                String scope = beanDefinition.getScope();
             
                printStream.printf("\t<bean id=\"%s\" class=\"%s\" scope=\"%s\"/>%n", beanName, beanClassName, scope);
            }
            
            printStream.println();
            
            printStream.println("</beans>");
            
            printStream.flush();
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
        finally {
            if (printStream != null) {
                printStream.close();
            }
            
            if (cacheFile.exists()) {
                makeReadWriteExecuteByAll(cacheFile);    
            }
        }

    }
    
}
