/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/spring/lazilyautowire/LazilyAutowired.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.spring.lazilyautowire;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * Notes that a method or field should be lazily autowired by the 
 * {@link LazilyAutowiredAnnotationBeanPostProcessor} when the
 * bean is instantiated.
 * 
 * For 'normal' autowiring, when a bean is instantiated, beans are retrieved
 * for all of the member autowired fields and methods.  This effectively
 * builds up a large bean tree with all dependencies satisfied.
 * 
 * For 'lazy' autowiring, when a bean is instantiated, proxies are created
 * for all of the member lazy autowired fields and methods.  Then when
 * the proxy is accessed for the first time, the appropriate bean is
 * looked up from the bean factory. This allows for 'delaying' the
 * instantiation of lazily initialized beans until first use
 * (which may never occur in a particular CLI invocation).
 * 
 * LazilyAutowiring cannot be used for primitives or arrays.
 * 
 * LazilyAutowiring requires the use of a {@link LazilyAutowiredAnnotationBeanPostProcessor}
 * in the bean context.
 * 
 * Advantages over normal autowiring:
 * <OL>
 * <LI>lazily initialized beans will not be instantiated until first use</LI>
 * </OL>
 * 
 * Disadvantages:
 * <OL>
 * <LI>unsatisfied dependencies will not be discovered until the proxy is accessed for the first time.</LI>
 * <LI>you cannot check the autowired property for null (the proxy will always be non-null)</LI>
 * </OL>
 * 
 * @author philipc
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
@Retention(RetentionPolicy.RUNTIME)
@Target({ElementType.METHOD, ElementType.FIELD})
public @interface LazilyAutowired {
    
    /**
     * True if you would like the reference to the proxy to be replaced
     * by a reference to the target bean when the target bean is resolved. 
     */
    boolean replacementAllowed() default true;
}
