/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/spring/lazilyautowire/LazilyAutowiredAnnotationBeanPostProcessor.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.spring.lazilyautowire;

import java.beans.PropertyDescriptor;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Proxy;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import net.sf.cglib.proxy.Enhancer;

import org.springframework.beans.BeanUtils;
import org.springframework.beans.PropertyValues;
import org.springframework.beans.factory.BeanCreationException;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.BeanFactoryAware;
import org.springframework.beans.factory.annotation.InjectionMetadata;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.beans.factory.config.DependencyDescriptor;
import org.springframework.beans.factory.config.InstantiationAwareBeanPostProcessorAdapter;
import org.springframework.core.MethodParameter;
import org.springframework.util.ReflectionUtils;

import com.netapp.common.logging.NALogger;

/**
 * Adds support for the {@link LazilyAutowired} annotation.
 * 
 * Creates proxies for all lazily autowired fields and method parameters.
 * 
 * @author philipc
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class LazilyAutowiredAnnotationBeanPostProcessor extends InstantiationAwareBeanPostProcessorAdapter implements BeanFactoryAware {
    
    private static final NALogger logger = NALogger.getLogger();

    private final Map<Class<?>, InjectionMetadata> injectionMetadataCache =
        new ConcurrentHashMap<Class<?>, InjectionMetadata>();

    private ConfigurableListableBeanFactory beanFactory;

    public void setBeanFactory(BeanFactory beanFactory) {
        if (!(beanFactory instanceof ConfigurableListableBeanFactory)) {
            throw new IllegalArgumentException(
                    "LazilyAutowiredAnnotationBeanPostProcessor requires a ConfigurableListableBeanFactory");
        }
        this.beanFactory = (ConfigurableListableBeanFactory) beanFactory;
    }
    
    public boolean postProcessAfterInstantiation(Object bean, String beanName) {
        InjectionMetadata metadata = findLazyAutowiringMetadata(bean.getClass());
        try {
            metadata.injectFields(bean, beanName);
        }
        // CHECKSTYLE:OFF
        catch (Throwable ex) {
            // CHECKSTYLE:ON
            throw new BeanCreationException(beanName, "Lazy autowiring of fields failed", ex);
        }
        return true;
    }

    public PropertyValues postProcessPropertyValues(
            PropertyValues pvs, PropertyDescriptor[] pds, Object bean, String beanName) {

        InjectionMetadata metadata = findLazyAutowiringMetadata(bean.getClass());
        try {
            metadata.injectMethods(bean, beanName, pvs);
        }
        // CHECKSTYLE:OFF
        catch (Throwable ex) {
            // CHECKSTYLE:ON
            throw new BeanCreationException(beanName, "Lazy autowiring of methods failed", ex);
        }
        return pvs;
    }


    private synchronized InjectionMetadata findLazyAutowiringMetadata(final Class clazz) {
        InjectionMetadata metadata = this.injectionMetadataCache.get(clazz);
        if (metadata == null) {
            final InjectionMetadata newMetadata = new InjectionMetadata();
            ReflectionUtils.doWithFields(clazz, new ReflectionUtils.FieldCallback() {
                public void doWith(Field field) {
                    LazilyAutowired annotation = field.getAnnotation(LazilyAutowired.class);
                    if (annotation != null) {
                        if (Modifier.isStatic(field.getModifiers())) {
                            throw new BeanCreationException("LazilyAutowired annotation is not supported on static fields: " + field);
                        }
                        if (field.getType().isArray()) {
                            throw new BeanCreationException("LazilyAutowired annotation cannot be used for an array: " + field);
                        }
                        if (field.getType().isPrimitive()) {
                            throw new BeanCreationException("LazilyAutowired annotation cannot be used for primitives: " + field);
                        }
                        newMetadata.addInjectedField(new LazyAutowiredElement(field, null, annotation.replacementAllowed()));
                    }
                }
            });
            ReflectionUtils.doWithMethods(clazz, new ReflectionUtils.MethodCallback() {
                public void doWith(Method method) {
                    LazilyAutowired annotation = method.getAnnotation(LazilyAutowired.class);
                    if (annotation != null) {
                        if (Modifier.isStatic(method.getModifiers())) {
                            throw new BeanCreationException("LazilyAutowired annotation is not supported on static methods");
                        }
                        if (method.getParameterTypes().length == 0) {
                            throw new BeanCreationException("LazilyAutowired annotation requires at least one argument: " + method);
                        }
                        
                        for (Class parameterType : method.getParameterTypes()) {
                            if (parameterType.isArray()) {
                                throw new BeanCreationException("LazilyAutowired annotation cannot be used for an array: " + method);
                            }
                            if (parameterType.isPrimitive()) {
                                throw new BeanCreationException("LazilyAutowired annotation cannot be used for primitives: " + method);
                            }
                        }
                        PropertyDescriptor pd = BeanUtils.findPropertyForMethod(method);
                        newMetadata.addInjectedMethod(new LazyAutowiredElement(method, pd, annotation.replacementAllowed()));
                    }
                }
            });
            metadata = newMetadata;
            this.injectionMetadataCache.put(clazz, metadata);
        }
        return metadata;
    }

    /**
     * Class representing injection information about an annotated field
     * or setter method.
     */
    private class LazyAutowiredElement extends InjectionMetadata.InjectedElement {

        private final boolean replacementAllowed;

        public LazyAutowiredElement(Member member, PropertyDescriptor pd, boolean replacementAllowed) {
            super(member, pd);
            this.replacementAllowed = replacementAllowed;
        }
        
        @Override
        protected void inject(Object bean, String beanName, PropertyValues propertyValues) throws Throwable {
            if (this.isField) {
                Field field = (Field) this.member;
                try {
                    
                    Object proxy = createProxy(beanName, bean, new DependencyDescriptor(field, true));
                    
                    ReflectionUtils.makeAccessible(field);
                    field.set(bean, proxy);
                }
                // CHECKSTYLE:OFF
                catch (Throwable ex) {
                    // CHECKSTYLE:ON
                    throw new BeanCreationException("Could not lazily autowire field: " + field, ex);
                }
            }
            else {
                if (this.pd != null && propertyValues != null && propertyValues.contains(this.pd.getName())) {
                    // Explicit value provided as part of the bean definition.
                    return;
                }
                Method method = (Method) this.member;
                Object[] arguments = new Object[method.getParameterTypes().length];
                try {

                    for (int i = 0; i < arguments.length; i++) {

                        arguments[i] = createProxy(beanName, bean, new DependencyDescriptor(new MethodParameter(method, i), true));
                        
                    }
                    
                    ReflectionUtils.makeAccessible(method);
                    method.invoke(bean, arguments);
                }
                // CHECKSTYLE:OFF
                catch (Throwable ex) {
                    // CHECKSTYLE:ON
                    throw new BeanCreationException("Could not lazily autowire method: " + method, ex);
                }
            }
        }

        private Object createProxy(String beanName, Object bean, DependencyDescriptor descriptor) {
            Class type = descriptor.getDependencyType();
            
            if (descriptor.getField() != null) {
                logger.internal("Creating lazy autowire proxy for field: " + descriptor.getField());
            }
            else {
                logger.internal("Creating lazy autowire proxy for method: "  + descriptor.getMethodParameter().getMethod() + " parameter: " + descriptor.getMethodParameter().getParameterIndex());
            }
            
            LazilyAutowiredInvocationHandler lazilyAutowiredInvocationHandler = new LazilyAutowiredInvocationHandler(
                LazilyAutowiredAnnotationBeanPostProcessor.this.beanFactory,
                beanName,
                (this.replacementAllowed) ? bean : null,
                descriptor);
            
            if (type.isInterface()) {
                return Proxy.newProxyInstance(
                    getClass().getClassLoader(),
                    new Class[] {type}, 
                    lazilyAutowiredInvocationHandler);
            }
            else {
                
                assertNoFinalMethodsExist(beanName, descriptor, descriptor.getDependencyType());
                
                Enhancer enhancer = new Enhancer();
                enhancer.setSuperclass(type);
                enhancer.setCallback(lazilyAutowiredInvocationHandler);
                
                return enhancer.create();
            }
        }

        private void assertNoFinalMethodsExist(String parentBeanName, DependencyDescriptor descriptor, Class type) {
            
            if (type == null || Object.class.equals(type)) {
                return;
            }
            
            for (Method method : type.getDeclaredMethods()) {
                if (Modifier.isFinal(method.getModifiers())) {
                    
                    if (descriptor.getField() != null) {
                        throw new BeanCreationException(
                            String.format(
                                "Classes to be LazilyAutowired must not have final methods.  "
                                + "The following method is final: %s.  "
                                + "An object of this class is being lazily autowired into the bean named %s in field %s",
                                method, parentBeanName, descriptor.getField()));
                    }
                    else {
                        throw new BeanCreationException(
                            String.format(
                                "Classes to be LazilyAutowired must not have final methods.  "
                                + "The following method is final: %s.  "
                                + "An object of this class is being lazily autowired into the bean named %s in method %s parameter %s",
                                method, parentBeanName, descriptor.getMethodParameter().getMethod(), descriptor.getMethodParameter().getParameterIndex()));
                    }
                    
                }
            }
            // recurse up the class hierarchy
            assertNoFinalMethodsExist(parentBeanName, descriptor, type.getSuperclass());
        }
    }
    
}
