/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/AdaptivePoller.java#1 $
 * 
 * Copyright (c) 2009 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import org.apache.commons.lang.time.DurationFormatUtils;

import com.netapp.common.logging.NALogger;

/**
 * Poll for completion of the {@link Pollee} using an adaptive polling interval.
 * 
 * The polling interval increases with the number of polling attempts.  If tuned correctly,
 * the ultimate effect is that:
 * <ul>
 * <li>Completion for short-running jobs will be recognized in a more responsive manner
 * <li>Polling frequency for long-running jobs will diminish over time, reducing load on 
 * the server and wasted cycles on the client   
 * </ul> 
 * 
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class AdaptivePoller {

    public static final ConfigProperty<Long> PROPERTY_MIN_SLEEP_TIME_IN_MILLIS = ConfigProperty.longProperty("poller.minSleepTimeInMillis", 1000L);
    public static final ConfigProperty<Long> PROPERTY_MAX_SLEEP_TIME_IN_MILLIS = ConfigProperty.longProperty("poller.maxSleepTimeInMillis", 30000L);
    public static final ConfigProperty<Double> PROPERTY_BACKOFF_FACTOR = new ConfigProperty<Double>(Double.class, "poller.backoffFactor", 1.03d);

    private static final NALogger logger = NALogger.getLogger(AdaptivePoller.class);
    
    private final long minSleepTimeInMillis;
    private final long maxSleepTimeInMillis;
    private final double backoffFactor;
    
    public AdaptivePoller(ConfigBundle configBundle) {
        this(
            PROPERTY_MIN_SLEEP_TIME_IN_MILLIS.lookup(configBundle),
            PROPERTY_MAX_SLEEP_TIME_IN_MILLIS.lookup(configBundle),
            PROPERTY_BACKOFF_FACTOR.lookup(configBundle)
        );
    }

    public AdaptivePoller(long sleepTimeInMillis) {
        this(sleepTimeInMillis, sleepTimeInMillis, 1.0d);
    }

    public AdaptivePoller(long minSleepTimeInMillis, long maxSleepTimeInMillis, double backoffFactor) {
        if (minSleepTimeInMillis <= 0) {
            throw new IllegalArgumentException(String.format("minSleepTimeInMillis (%d) must be greater than 0", minSleepTimeInMillis));
        }
        if (maxSleepTimeInMillis < minSleepTimeInMillis) {
            throw new IllegalArgumentException(String.format("maxSleepTimeInMillis (%d) must be greater than minSleepTimeInMillis (%d)", maxSleepTimeInMillis, minSleepTimeInMillis));
        }
        if (backoffFactor < 1.0d) {
            throw new IllegalArgumentException(String.format("backoffFactor (%f) must be greater than 1", backoffFactor));
        }
        this.minSleepTimeInMillis = minSleepTimeInMillis;
        this.maxSleepTimeInMillis = maxSleepTimeInMillis;
        this.backoffFactor = backoffFactor;
    }

    
    public void poll(Pollee pollee) {

        long sleepTime = this.minSleepTimeInMillis; 
        
        while (!pollee.isComplete()) {
            
            try {
                Thread.sleep(sleepTime);
            }
            catch (InterruptedException e) {
                // ignore and try again
            }
            
            sleepTime = Math.min(this.maxSleepTimeInMillis, (long) (sleepTime * this.backoffFactor));
            
            logger.internal("Polling status for job [%s] (adaptive poll interval at %s)...", 
                pollee, DurationFormatUtils.formatDurationHMS(sleepTime));
        }

        logger.internal("Finished polling for job [%s]", pollee);

    }
    
}
