/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/bootstrap/com/netapp/common/util/ClassUtils.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.io.File;
import java.io.StringReader;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Utility Class to handle manipulation of Class names and
 * other reflective activities.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public final class ClassUtils {
    /**
     * Regex pattern to check for Java Bean style method names.
     */
    private static final Pattern BEAN_METHOD_REGEX = Pattern.compile("(set|get|is)([A-Z].*)");
    
    /**
     * http://java.sun.com/docs/books/jls/third_edition/html/lexical.html#3.9
     */
    private static final Set<String> RESERVED_WORDS = new TreeSet<String>();
    static {
        RESERVED_WORDS.addAll(
            Arrays.asList("abstract",
                "assert",
                "boolean",
                "break",
                "byte",
                "case",
                "catch",
                "char",
                "class",
                "const",
                "continue",
                "default",
                "do",
                "double",
                "else",
                "enum",
                "extends",
                "final",
                "finally",
                "float",
                "for",
                "goto",
                "if",
                "implements",
                "import",
                "instanceof",
                "int",
                "interface",
                "long",
                "native",
                "new",
                "package",
                "private",
                "protected",
                "public",
                "return",
                "short",
                "static",
                "strictfp",
                "super",
                "switch",
                "synchronized",
                "this",
                "throw",
                "throws",
                "transient",
                "try",
                "void",
                "volatile",
                "while",
                /* These three aren't keywords but reserved literals */
                "true",
                "false",
                "null"));
    }
    
    private ClassUtils() {
        throw new RuntimeException("ClassUtils may not be instantiated");
    }

    /**
     * Check if the requested word is a Java keyword.
     * @param word
     * @return True if requested word is a Java keyword
     */
    public static boolean isJavaReservedWord(String word) {
        return RESERVED_WORDS.contains(word);
    }
    
    /**
     * Convert a class name to a file path
     * @param clazz
     * @return Relative file path of the supplied class or package, excluding extension
     */
    public static String classNameToFilePath(String clazz) {
        return clazz.replace('.', File.separatorChar);
    }

    /**
     * Extract the package name from a qualified class name, or null if the
     * class is in the base package.
     * @param clazz
     * @return package name of the supplied class or null if the class is in the default package
     */
    public static String extractPackageName(String clazz) {
        int last = clazz.lastIndexOf(".");
        if (last >= 0) {
            return clazz.substring(0, last);
        }
        return null;
    }

    /**
     * Extract the unqualified class name from a qualified class name
     * @param clazz
     * @return unqualified class name of the supplied class
     */
    public static String extractUnqualifiedClassName(String clazz) {
        int last = clazz.lastIndexOf(".");
        if (last >= 0) {
            return clazz.substring(last + 1);
        }
        /* There was no package name, so this is the class name */
        return clazz;
    }
    
    /**
     * Returns the unqualified class name of the specified object's class.
     * @param o
     * @return unqualified class name of the specified object's class
     */
    public static String unqualifiedClassNameForObject(Object o) {
        String className = o.getClass().getName();
        return extractUnqualifiedClassName(className);
    }

    /**
     * Check if the supplied method is a Java Bean accessor or setter.
     * A method is a bean method if:
     * <ul>
     *   <li> The method starts with "get" or "is" and has no arguments; or
     *   <li> The method starts with "set" and has one argument
     * </ul>
     * @param method The method to check
     * @return true if this method appears to be a bean method, false otherwise
     */
    public static boolean isBeanMethod(Method method) {
        Matcher matcher = BEAN_METHOD_REGEX.matcher(method.getName());

        // Setters should have exactly 1 argument
        // Matchers should have none
        if (matcher.matches()) {
            if ("set".equals(matcher.group(1)) && (method.getParameterTypes().length == 1)) {
                return true;
            }
            else if (method.getParameterTypes().length == 0) {
                return true;
            }
        }
        return false;
    }

    /**
     * Return a list of methods on an obect that appear to be Java Bean setters for a given field
     * @param clazz Class to look for method
     * @param fieldName Java Bean field name
     * @return List of methods (if any) that appear to be valid Java Bean setters
     */
    public static List<Method> findSetterMethod(Class<?> clazz, String fieldName) {
        List<Method> methods = new ArrayList<Method>();

        String methodName = "set" + StringUtils.initialUpperCase(fieldName);

        for (Method m : clazz.getMethods()) {
            if (m.getName().equals(methodName) && (m.getGenericParameterTypes().length == 1)) {
                methods.add(m);
            }
        }
        return methods;
    }

    /**
     * Return the Java Bean getter for a field
     * @param clazz Class to look for method
     * @param fieldName Java Bean field name
     * @return Getter method for field
     */
    public static Method findGetterMethod(Class<?> clazz, String fieldName) {
        String getMethodName = "get" + StringUtils.initialUpperCase(fieldName);
        String isMethodName = "is" + StringUtils.initialUpperCase(fieldName);
        for (Method m : clazz.getMethods()) {
            if ((m.getName().equals(isMethodName) || m.getName().equals(getMethodName))
                && (m.getGenericParameterTypes().length == 0)) {
                return m;
            }
        }
        return null;
    }

    /**
     * Convert a JavaBeans accessor method to a Java field name
     * @param method The method to retrieve the field from
     * @return The field the supplied method acts on
     */
    public static String extractFieldNameFromBeanMethod(Method method) {
        if (!isBeanMethod(method)) {
            throw new IllegalArgumentException("Method " + method.getName() + " is not a JavaBean method");
        }
        Matcher m = BEAN_METHOD_REGEX.matcher(method.getName());
        m.matches();
        return StringUtils.initialLowerCase(m.group(2));
    }

    /**
     * Determine the Type of the field a bean method acts on.
     * This is determined:
     * <ul> 
     *   <li> For get or is methods by the return type
     *   <li> For set methods by the type of the argument
     * </ul>
     * @param method The method to retrieve the type from
     * @return The Type of the field the supplied method acts on
     */
    public static Type extractFieldTypeFromBeanMethod(Method method) {
        if (!isBeanMethod(method)) {
            throw new IllegalArgumentException("Method " + method.getName() + " is not a JavaBean method");
        }

        if (method.getName().startsWith("set")) {
            return method.getGenericParameterTypes()[0];
        }
        else {
            return method.getGenericReturnType();
        }
    }

    /**
     * Is the supplied Type a generic collection?
     * @param t
     * @return True if T is a subclass of Collection<E>, false otherwise
     */
    public static boolean isGenericCollection(Type t) {
        if (t instanceof ParameterizedType) {
            ParameterizedType pt = (ParameterizedType) t;
            if (pt.getRawType() instanceof Class) {
                Class<?> rawType = (Class<?>) pt.getRawType();
                return Collection.class.isAssignableFrom(rawType);
            }
        }
        return false;
    }

    /**
     * Extract the base class from the supplied type.
     * If Type is a class, the result is the class itself.
     * If Type is a collection, the result is the collection type
     * @param t
     * @return The immediate type of the supplied type
     */
    public static Class<?> extractImmediateClass(Type t) {
        Class<?> clazz = null;
        if (t instanceof Class) {
            clazz = (Class<?>) t;
        }
        else if (isGenericCollection(t)) {
            clazz = extractCollectionClass(t);
        }

        if (clazz == null) {
            throw new IllegalArgumentException("Type " + t + " is not supported");
        }

        return clazz;

    }

    /**
     * Extract the Object Class from the supplied Type.
     * If Type is a class, the result is the class itself.
     * If Type is a collection, the result is the element type
     * @return The class of the object represented or stored in t
     *
     */
    public static Class<?> extractObjectClass(Type t) {
        Class<?> clazz = null;
        if (t instanceof Class) {
            clazz = (Class<?>) t;
        }
        else if (isGenericCollection(t)) {
            ParameterizedType pt = (ParameterizedType) t;
            // This should work, since Collection must have one type
            Type parameterType = pt.getActualTypeArguments()[0];
            if (parameterType instanceof Class) {
                clazz = (Class<?>) parameterType;
            }
        }

        if (clazz == null) {
            throw new IllegalArgumentException("Type " + t + " is not supported");
        }

        return clazz;
    }

    /**
     * Extract the Collection class from the supplied Type.
     * @param t
     * @return The class T implements that extends Collection, or null.
     */
    public static Class<?> extractCollectionClass(Type t) {
        Class<?> clazz = null;
        if (isGenericCollection(t)) {
            ParameterizedType pt = (ParameterizedType) t;
            Type ct = pt.getRawType();
            clazz = (Class<?>) ct;
        }
        return clazz;
    }

    /**
     * Attempt to create an object of the supplied type with the supplied content.
     * This proceeeds as follows: <ul>
     * <li>If clazz is a string, content is directly returned.
     * <li>If clazz is a primitive or one of the primitive encapsulation classes, the appropriate parse is called and returned.
     * <li>If clazz has a constructor that takes a single string value, one is instantiated reflectively with that constructor.
     * </ul>
     * Under any other circumstances, an IllegalArgumentException is thrown.
     * @param clazz Class (or primitive Class) of the new object to create
     * @param content Initial value of the new object
     * @return new object as defined in the method description
     */
    @SuppressWarnings("unchecked")
    public static <T> T createObject(Class<T> clazz, String content) {
        if (content == null) {
            return null;
        }

        if (!clazz.isPrimitive()
            && (Modifier.isInterface(clazz.getModifiers()) || Modifier.isAbstract(clazz.getModifiers()))) {
            throw new IllegalArgumentException(clazz.getName() + " may not be directly instantiated");
        }

        Object result = null;
        Throwable t = null;

        try {
            if (clazz.equals(String.class)) {
                result = content;
            }
            else if (clazz.isEnum()) {
                result = Enum.valueOf(clazz.asSubclass(Enum.class), content.trim());
            }
            else if (clazz.equals(Byte.class) || clazz.equals(Byte.TYPE)) {
                // We have to do this so the cast below works.
                // Attempting to do Byte.TYPE.cast(result) will throw a cast exception
                // but because of erasure, if T == Byte.TYPE, we can replace it with Byte.class
                // without throwing an exception, and the JVM will rebox as needed.
                clazz = (Class<T>) Byte.class;
                result = Byte.parseByte(content.trim());
            }
            else if (clazz.equals(Short.class) || clazz.equals(Short.TYPE)) {
                clazz = (Class<T>) Short.class;
                result = Short.parseShort(content.trim());
            }
            else if (clazz.equals(Integer.class) || clazz.equals(Integer.TYPE)) {
                clazz = (Class<T>) Integer.class;
                result = Integer.parseInt(content.trim());
            }
            else if (clazz.equals(Long.class) || clazz.equals(Long.TYPE)) {
                clazz = (Class<T>) Long.class;
                result = Long.parseLong(content.trim());
            }
            else if (clazz.equals(Boolean.class) || clazz.equals(Boolean.TYPE)) {
                clazz = (Class<T>) Boolean.class;
                content = content.trim();
                // Do our own comparison instead of using Boolean.parse() and
                // throw an illegal argument exception if it's neither. The
                // Java standard is more lax (anything not TRUE is false.)
                // burt 317418
                if ("true".equalsIgnoreCase(content)) {
                    result = Boolean.TRUE;
                }
                else if ("false".equalsIgnoreCase(content)) {
                    result = Boolean.FALSE;
                }
                else {
                    throw new IllegalArgumentException("Invalid boolean value " + content + "; must be TRUE or FALSE");
                }
            }
            else if (clazz.equals(Float.class) || clazz.equals(Float.TYPE)) {
                clazz = (Class<T>) Float.class;
                result = Float.parseFloat(content.trim());
            }
            else if (clazz.equals(Double.class) || clazz.equals(Double.TYPE)) {
                clazz = (Class<T>) Double.class;
                result = Double.parseDouble(content.trim());
            }
            // Dates are always returned as the number of seconds from epoch from ZAPI.
            // Using this in non-zapi contexts inherits the same convention for now.
            // Date(Long) takes milliseconds since epoch
            else if (clazz.equals(Date.class)) {
                long l = Long.parseLong(content.trim()) * 1000;
                result = new Date(l);
            }
            // Attempt to use the single-string argument constructor
            else {
                Constructor con = clazz.getConstructor(String.class);
                result = con.newInstance(content);
            }
        }
        catch (NumberFormatException e) {
            throw new IllegalArgumentException("Number format exception while converting " + content + " to "
                + clazz.getSimpleName(), e);
        }
        catch (NoSuchMethodException e) {
            throw new IllegalArgumentException("No suitable constructor for " + clazz.getName() + "found", e);
        }
        catch (InstantiationException e) {
            t = e;
        }
        catch (IllegalAccessException e) {
            t = e;
        }
        catch (InvocationTargetException e) {
            t = e;
        }

        if (t != null) {
            throw new IllegalArgumentException("Reflection exception while instantiating " + clazz.getName() + "("
                + content + ")", t);
        }
        return clazz.cast(result);
    }

    /**
     * Create an object of the given class using the default constructor.
     * 
     * @throws IllegalArgumentException if the class has no default constructor
     */
    public static <T> T createObject(Class<T> clazz) {
        try {
            Constructor<T> defaultConstructor = clazz.getDeclaredConstructor();
            defaultConstructor.setAccessible(true);
            return defaultConstructor.newInstance();
        }
        catch (SecurityException e) {
            throw new IllegalArgumentException(e);
        }
        catch (NoSuchMethodException e) {
            throw new IllegalArgumentException(e);
        }
        catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(e);
        }
        catch (InstantiationException e) {
            throw new IllegalArgumentException(e);
        }
        catch (IllegalAccessException e) {
            throw new IllegalArgumentException(e);
        }
        catch (InvocationTargetException e) {
            throw new IllegalArgumentException(e);   
        }

    }
    
    /**
     * Create a new collection with the content as the first argument. If the class is actually an interface, 
     * create a new base, concrete type. The interfaces supported are List, Set and Queue.
     * 
     * The generic type associated with each of the data types is expected to be Object.
     * 
     * @param <T>
     * @param clazz
     * @param content
     * @return Appropriate implementation of interface <code>clazz</code>
     */
    @SuppressWarnings("unchecked")
    public static <T> T createCollection(Class<T> clazz, String content) {
        if (clazz.isPrimitive()) {
            return createObject(clazz, content);
        }

        Collection<Object> result = null;
        Throwable t = null;

        try {
            if (typeAssignableFrom(List.class, clazz)) {
                if (clazz.isInterface()) {
                    result = new ArrayList<Object>();
                }
                else {
                    result = (Collection<Object>) clazz.newInstance();
                }
            }
            else if (typeAssignableFrom(Set.class, clazz)) {
                if (clazz.isInterface()) {
                    // Changed from TreeSet because TreeSet assumes that all elements being added implement Comparable
                    // which isn't a valid restriction to be placing on the workflow sets.
                    result = new HashSet<Object>();
                }
                else {
                    result = (Collection<Object>) clazz.newInstance();
                }
            }
            else if (typeAssignableFrom(Queue.class, clazz)) {
                if (clazz.isInterface()) {
                    result = new LinkedList<Object>();
                }
                else {
                    result = (Collection<Object>) clazz.newInstance();
                }
            }
            else {
                throw new IllegalArgumentException("Unsupported collection type " + clazz.getName());
            }

            if (null != content) {
                result.add(content);
            }
        }
        catch (InstantiationException e) {
            t = e;
        }
        catch (IllegalAccessException e) {
            t = e;
        }

        if (t != null) {
            throw new IllegalArgumentException("Reflection exception while instantiating " + clazz.getName() + "("
                + content + ")", t);
        }

        return clazz.cast(result);
    }

    public static void setBeanProperty(Object bean, String property, String value) {
        List<Method> setters = ClassUtils.findSetterMethod(bean.getClass(), property);
        if (setters.size() != 1) {
            // TODO : We need to move this out of bootstrap so I can use the logger
            // and throw better exceptions
            throw new IllegalArgumentException("No property " + property + " on " + bean.getClass().getName());
        }
        try {
            Method m = setters.get(0);
            Object o = ClassUtils.createObject(m.getParameterTypes()[0], value);
            m.invoke(bean, o);
        }
        catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(e);
        }
        catch (IllegalAccessException e) {
            throw new IllegalArgumentException(e);
        }
        catch (InvocationTargetException e) {
            throw new IllegalArgumentException(e);
        }
    }

    /**
     * Check if source may be assigned to target.
     * This needs to be updated with better rules for generics.
     * @param target
     * @param source
     * @return true if source may be assigned to target
     */
    public static boolean typeAssignableFrom(Type target, Type source) {
        // Everything can be assigned to Object
        if (target.equals(Object.class)) {
            return true;
        }

        // All parameterized types can be assigned if exactly equal
        if (target.equals(source)) {
            return true;
        }

        if ((target instanceof Class) && (source instanceof Class)
            && (((Class<?>) target).isAssignableFrom((Class<?>) source))) {
            return true;
        }
        
        /*
         * Check for base types and auto-boxing
         */
        if ((target.equals(Byte.class) || target.equals(Byte.TYPE)) 
                && (source.equals(Byte.class) || source.equals(Byte.TYPE))) {
            return true;
        }
        else if ((target.equals(Short.class) || target.equals(Short.TYPE)) 
                && (source.equals(Short.class) || source.equals(Short.TYPE))) {
            return true;
        }
        else if ((target.equals(Integer.class) || target.equals(Integer.TYPE)) 
                && (source.equals(Integer.class) || source.equals(Integer.TYPE))) {
            return true;
        }
        else if ((target.equals(Long.class) || target.equals(Long.TYPE)) 
                && (source.equals(Long.class) || source.equals(Long.TYPE))) {
            return true;
        }
        else if ((target.equals(Boolean.class) || target.equals(Boolean.TYPE)) 
                && (source.equals(Boolean.class) || source.equals(Boolean.TYPE))) {
            return true;
        }
        else if ((target.equals(Float.class) || target.equals(Float.TYPE)) 
                && (source.equals(Float.class) || source.equals(Float.TYPE))) {
            return true;
        }
        else if ((target.equals(Double.class) || target.equals(Double.TYPE)) 
                && (source.equals(Double.class) || source.equals(Double.TYPE))) {
            return true;
        }
        
        return false;
    }

    /**
     * Check if the source is assignable to any of the supported Collection interfaces.
     * 
     * @param source
     * @return <code>true</code> if source is a supported Collection interface
     */
    public static boolean typeAssignableFromCollection(Type source) {
        if (typeAssignableFrom(List.class, source) || typeAssignableFrom(Set.class, source)
            || typeAssignableFrom(Queue.class, source)) {
            return true;
        }
        else if (typeAssignableFrom(List.class, source.getClass()) || typeAssignableFrom(Set.class, source.getClass())
                || typeAssignableFrom(Queue.class, source.getClass())) {
            return true;
        }

        return false;
    }

    /**
     * Create an array of Class objects from a given array of Objects
     */
    public static Class<?>[] extractClasses(Object... objects) {
        
        Class<?>[] classes = new Class[objects.length];
        
        for (int i = 0; i < objects.length; i++) {
            classes[i] = objects[i].getClass();
        }
        
        return classes;
    }
    
    /**
     * Convert a string representing a Java type declaration into generic types.
     * Currently handles simple classes, parameterized types (List&lt;String&gt;)
     * and upper-bounded wildcards (List&lt;? extends Interface&gt;).
     * 
     * Classes in java.lang will be resolved without prefix.
     * @param declaration
     * @return Type declaration of the supplied string
     * @throws ClassNotFoundException
     * @throws IllegalArgumentException
     */
    public static Type parseTypeDeclaration(String declaration) throws ClassNotFoundException {
        JavaTypeParser parser = new JavaTypeParser(new StringReader(declaration));
        try {
            Type type = parser.Type();
            
            if (type instanceof ParameterizedTypeImpl) {
                validateTypeParameters((ParameterizedTypeImpl)type);
            }
            
            return type;
        }
        catch (ParseException e) {
            throw new IllegalArgumentException(e.getMessage(), e);
        }
    }

    private static void validateTypeParameters(ParameterizedTypeImpl type) {
        if (!(type.getRawType() instanceof Class)) {
            throw new IllegalArgumentException("Parameterized type " + type.getRawType() + " is not a parameterized type.");
        }
        
        Class<?> clazz = (Class<?>)type.getRawType();
        if (type.getActualTypeArguments().length != clazz.getTypeParameters().length) {
            throw new IllegalArgumentException("Invalid number of type arguments for " + clazz.getName());
        }
        
        for (Type typeArgument : type.getActualTypeArguments()) {
            if (typeArgument instanceof ParameterizedTypeImpl) {
                validateTypeParameters((ParameterizedTypeImpl)typeArgument);
            }
        }
    }

    static Class<?> resolveClass(String name) throws ClassNotFoundException {
        Class<?> clazz = null;
        try {
            clazz = Class.forName(name);
        }
        catch (ClassNotFoundException e) {
        }

        if (clazz == null && !name.contains(".")) {
            try {
                clazz = Class.forName("java.lang." + name);
            }
            catch (ClassNotFoundException e) {
            }
        }

        if (clazz == null) {
            throw new ClassNotFoundException(name);
        }
        return clazz;
    }
    
    
}
