/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/parser/Block.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.parser;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Represents a comment block parsed into Tags.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
class Block {
    private static final NALogger logger = NALogger.getLogger(Block.class);
    
    /** Pattern matching an opening zapidoc comment */
    static final Pattern BEGIN_BLOCK = Pattern.compile("\\s*/\\*+\\s*");
    
    /** Pattern matching the first line of a tag */
    static final Pattern BEGIN_TAG = Pattern.compile("\\s*\\*+\\s*@(\\w+)\\s+(\\S.*)");
    
    /** Pattern matching a tag continuation */
    static final Pattern CONTINUE_TAG = Pattern.compile("\\s*\\**\\s*([^\\s\\*@].*)");
    
    /** Pattern matching a closing zapidoc comment */
    static final Pattern END_BLOCK = Pattern.compile("\\s*\\*+\\/.*");
    
    /** The list of tags in this block */
    private Queue<Tag> tags;
    
    /**
     * Construct a new block.
     * @see #readBlocks
     */
    private Block() {
        tags = new LinkedList<Tag>();
    }
    
    /**
     * Add a new tag to this block
     * @param tag new Tag to add
     */
    private void addTag(Tag tag) {
        this.tags.add(tag);
    }
    
    /**
     * Retrieve the list of tags on this block
     * @return tags declared in this block
     */
    Queue<Tag> getTags() {
        return this.tags;
    }   
    
    /**
     * Read a list of blocks from the supplied reader
     * @param reader Input to read from
     * @return List of blocks read
     * @throws IOException An IO error was encountered
     */
    static List<Block> readBlocks(BufferedReader reader) throws IOException {
        List<Block> blocks = new ArrayList<Block>();
        
        Block block = readBlock(reader);
        while (block != null) {
            blocks.add(block);
            block = readBlock(reader);
        }
        return blocks;
        
    }
    
    /**
     * Read an individual block from the supplied input stream.
     * Consumes any noise prior to an opening comment.
     * @param reader Input to read from
     * @return New block, or null if the end of the stream was reached with no block parsed
     * @throws IOException An IO error was encountered
     */
    private static Block readBlock(BufferedReader reader) throws IOException {
        Block block = null;
        Tag tag = null;
        String line = reader.readLine();
        while (line != null) {
            line = line.replaceAll("\\s+", " ");
            Matcher blockBegin = BEGIN_BLOCK.matcher(line);
            Matcher tagBegin = BEGIN_TAG.matcher(line);
            Matcher tagCont = CONTINUE_TAG.matcher(line);
            Matcher blockEnd = END_BLOCK.matcher(line);
            
            if (blockBegin.matches()) {
                assert block == null;
                block = new Block();
            }
            else if (blockEnd.matches()) {
                break;
            }
            else if (tagBegin.matches()) {
                tag = null;
                try {
                    if (block != null) {
                        tag = new Tag(tagBegin.group(1), tagBegin.group(2));
                        block.addTag(tag);
                    }
                }
                catch (IllegalArgumentException e) {
                    logger.debug(MsgKey.ZAPI_GENERATOR_UNKNOWN_TAG, tagBegin.group(1));
                }
            }
            else if (tagCont.matches()) {
                if (tag != null) {
                    tag.append(tagCont.group(1));
                }
            }
            line = reader.readLine();
        }
        return block;
    }
}